<!DOCTYPE gui SYSTEM "kpartgui.dtd">
<gui name="kbackup">
<MenuBar>
  <Menu name="file"><text>&amp;File</text>
    <Action name="newProfile"/>
    <Action name="loadProfile"/>
    <Action name="recentProfiles"/>
    <Action name="saveProfile"/>
    <Action name="saveProfileAs"/>
    <Action name="profileSettings"/>
  </Menu>
  <Menu name="settings"><text>&amp;Settings</text>
    <Action name="dockInSysTray"/>
    <Action name="enableAllMessages"/>
    <Action name="showHiddenFiles"/>
  </Menu>
</MenuBar>
</gui>
